#include <stdio.h>
#include "llc_iface.h"
#include "firewire.h"

//
// Link-layer interface to the TSB12LV32 chip,
// via a small hardware wrapper on fpga1.
// 

// Descriptors used to access TM4 ports
static int	p_signals_out, p_signals_in;
static int	initialized = 0;

// Two ports are used for everything: signals_in and signals_out.
// 
// The signals_in port tells the hardware to
// either perform a read or a write, by sending
// a 32-bit command word:
// bit 32: 1 = write, 0 = read
// bits 0-6: register address
// bits 8-24: 16-bit write data (for writes)
//
// 16-bit read data is returned in a padded 32-bit
// word on the signals_out port


// Initialize the interface to the TM4 Ports Package
// and open the signals_in and signals_out ports.
int llc_init()
{
	tm_init("");
	
	p_signals_in = tm_open("signals_in", "w");
	p_signals_out = tm_open("signals_out", "r");

	if (p_signals_in < 0 || p_signals_out < 0)
		return -1;

	initialized = 1;
	return 0;
}

// Read a 16-bit word from a register on the link chip,
// given its 7-bit address.
int llc_read(int addr)
{
	int cmd = addr;
	int result;
	
	if (!initialized)
		return -1;
	
	// Issue a read command
	if (tm_write(p_signals_in, (char*)&cmd, 4) != 4)
		return -1;

	// Retrieve data
	if (tm_read(p_signals_out, (char*)&result, 4) != 4)
		return -1;

	return result;
}

// Write to a register on the link chip.
int llc_write(int addr, int data)
{
	int diag_reg;
	int cmd = 0x80000000 | ((data & 0xFFFF) << 8) | addr;

	if (!initialized)
		return -1;

	// Wait until the link chip is ready to accept
	// a write, by polling the 0x20 diagnostic register.
	do
	{
		diag_reg = llc_read(0x20);
	}
	while (diag_reg != -1 && (diag_reg & 0xF000));

	if (diag_reg == -1)
		return -1;
	
	// Issue a write command, along with data
	if (tm_write(p_signals_in, (char*)&cmd, 4) != 4)
		return -1;

	return 0;
}

// Reads a 32-bit word from two adjacent 16-bit
// registers. Address should be 4-aligned.
int llc_read32(int addr, int* buf)
{
	int readhi, readlo;

	if (addr & 0x3)
		return -1;

	if ((readhi = llc_read(addr)) < 0)
		return -1;
	if ((readlo = llc_read(addr + 2)) < 0)
		return -1;

	*buf = (readhi << 16) | readlo;
	return 0;
}

// Write a 32-bit value to two consecutive
// 16-bit registers. Address should be 4-aligned.
int llc_write32(int addr, int data)
{
	if (addr & 0x3)
		return -1;

	// Upper 16-bits go to lower address first
	if (llc_write(addr, data >> 16) < 0)
		return -1;

	if (llc_write(addr + 2, data & 0xFFFF) < 0)
		return -1;

	return 0;
}

// Write a 16-bit value to a register, but only
// change certain bits of the target register,
// wherever the mask is '1'.
int llc_write_masked(int addr, int data, int mask)
{
	int data_orig = llc_read(addr);
	int data_new;
	
	if (data_orig < 0)
		return -1;;

	data_new = (data_orig & (~mask)) | (data & mask);

	if (llc_write(addr, data_new) < 0)
		return -1;

	return 0;
}

// Send a firewire data packet to another node on the bus.
// The type of packet depends on the "tcode" field of the
// llc_packet structure, and this also determines which
// other fields are actually going to be valid/useful.
int llc_send_packet(llc_packet* pkt)
{
	int i;

	// The first quadlet of a packet is written to 0x50-52
	// Each successive middle quadlet is written to 0x54-56
	// The last quadlet of the packet is written to 0x58-5a
	
	// The header quadlet is common to all packet types.
	llc_write32(0x50, (pkt->tcode<<4) | (pkt->rt<<8) | (pkt->tlabel<<10) | (pkt->spd<<16));

	switch(pkt->tcode)
	{
		case TC_REQ_QUAD_WRITE:
			llc_write(0x54, pkt->dest_id);
			llc_write(0x56, pkt->dest_offset >> 32LL);
			llc_write32(0x54, pkt->dest_offset & 0xFFFFFFFF);
			llc_write32(0x58, *pkt->data);
		break;

		case TC_REQ_BLOCK_WRITE:
		case TC_REQ_BLOCK_READ:
		case TC_REQ_LOCK:
			llc_write(0x54, pkt->dest_id);
			llc_write(0x56, pkt->dest_offset >> 32LL);
			llc_write32(0x54, pkt->dest_offset & 0xFFFFFFFF);
			if (pkt->tcode == TC_REQ_LOCK)
			{
				llc_write(0x54, pkt->data_length);
				llc_write(0x56, pkt->ext_tcode);
				for (i = 0; i < pkt->data_length - 1; i++)
					llc_write32(0x54, pkt->data[i]);
				llc_write32(0x58, pkt->data[i]);
			}
			if (pkt->tcode == TC_REQ_BLOCK_WRITE)
			{
				llc_write(0x54, pkt->data_length);
				llc_write(0x56, 0x0000);
				for (i = 0; i < pkt->data_length - 1; i++)
					llc_write32(0x54, pkt->data[i]);
				llc_write32(0x58, pkt->data[i]);
			}
			else
			{
				llc_write(0x58, pkt->data_length);
				llc_write(0x5a, 0x0000);
			}
		break;
				
		case TC_RESP_WRITE:
			llc_write(0x58, pkt->dest_id);
			llc_write(0x5a, pkt->rcode << 12);
		break;

		case TC_REQ_QUAD_READ:
			llc_write(0x54, pkt->dest_id);
			llc_write(0x56, pkt->dest_offset >> 32LL);
			llc_write32(0x58, pkt->dest_offset & 0xFFFFFFFF);
		break;

		case TC_RESP_QUAD_READ:
			llc_write(0x54, pkt->dest_id);
			llc_write(0x56, pkt->rcode << 12);
			llc_write32(0x58, *pkt->data);
		break;

		case TC_RESP_LOCK:
		case TC_RESP_BLOCK_READ:
			llc_write(0x54, pkt->dest_id);
			llc_write(0x56, pkt->rcode << 12);
			llc_write32(0x54, 0L);
			llc_write(0x54, pkt->data_length);
			llc_write(0x56, pkt->ext_tcode);
			for (i = 0; i < pkt->data_length - 1; i++)
				llc_write32(0x54, pkt->data[i]);
			llc_write32(0x58, pkt->data[i]);
		break;
	}


	return 0;
}

// Free a packet structure from memory
void llc_free_packet(llc_packet* pkt)
{
	if (pkt->data != 0)
		free(pkt->data);

	free(pkt);
}

// Allocates space for and receives a packet from
// another node on the firewire bus.
// No checking is done to guarantee that a packet
// is actually available for reception!
llc_packet*	llc_recv_packet()
{
	int data;
	int i;
	llc_packet* pkt;

	// Allocate/initialize packet structure
	pkt = (llc_packet*) malloc(sizeof(llc_packet));
	memset (pkt, 0, sizeof(llc_packet));
	
	// Get the token-quadlet, which is the very
	// first quadlet in the packet. It gives the
	// size of the actual packet along with its
	// transmission speed and acknowledgement code.
	pkt->token.quadlet_count = llc_read(0x60);
	data = llc_read(0x62);
	pkt->token.speed = (data>>4) & 0x3;
	pkt->spd = pkt->token.speed;
        pkt->token.ack_code = (data>>8) & 0x1F;

	// Extract data from the 2nd quadlet
	pkt->dest_id = llc_read(0x60);
	data = llc_read(0x62);
	pkt->tcode = (data>>4) & 0xF;
	pkt->rt = (data>>8) & 0x3;
	pkt->tlabel = (data>>10) & 0x3F;
	
	// Extract data fromt he 3rd quadlet
	pkt->source_id = llc_read(0x60);
	data = llc_read(0x62);
	pkt->rcode = data >> 12;
	pkt->dest_offset = (long long int)data << 32LL;
	llc_read32(0x60, &data);
	pkt->dest_offset |= data;

	// What comes next depends on the packet's
	// transaction code (tcode) which determines
	// its type.
	switch (pkt->tcode)
	{
		// Quadlet read-responses or write-requests
		// come with 1 quadlet of data attached.
		case TC_REQ_QUAD_WRITE:
		case TC_RESP_QUAD_READ:
			pkt->data_length = 1;
			break;
		
		// Block transfers come with multiple quadlets,
		// specified in an extra packet field.
		case TC_REQ_BLOCK_WRITE:
		case TC_RESP_BLOCK_READ:
			pkt->data_length = llc_read(0x60);
			pkt->ext_tcode = llc_read(0x62);

		// Assume any remaining quadlets are data.
		default:
			pkt->data_length = pkt->token.quadlet_count - 3;
	}
	
	// Allocate space for data (could be 0 bytes if none)
	// and read it in from the remainder of the packet.
	pkt->data = (int*) malloc(pkt->data_length * sizeof(int));
	for (i = 0; i < pkt->data_length; i++)
		llc_read32(0x60, &pkt->data[i]);

	return pkt;
}

// Allocate space for and receive an array of
// firewire self-id packets. This functino does
// not do any checks to see if there are actually
// any self-id packets ready for reception!
//
// It is assumed that "full self-id packet mode"
// is turned off, such that the CRCs for each
// self-id quadlet are NOT present in the receive buffer.
llc_selfid_packet*	llc_recv_selfid_packets()
{
	int data;
	llc_packet_token token;
	int i;
	llc_selfid_packet* pkts;

	// Receive and parse token quadlet
	token.quadlet_count = llc_read(0x60);
	data = llc_read(0x62);
	token.speed = (data>>4) & 0x3;
	token.ack_code = (data>>8) & 0xF;
	
	// Knowing how many self-id packets there are,
	// allocate space for the array.
	pkts = (llc_selfid_packet*) malloc(token.quadlet_count * sizeof(llc_selfid_packet));
	
	// Retrieve and parse each self-id packet, which is
	// 1 quadlet long.
	for (i = 0; i < token.quadlet_count; i++)
	{
		pkts[i].num_pkts = token.quadlet_count;

		data = llc_read(0x60);
		pkts[i].phy_id = (data>>8) & 0x3F;
		pkts[i].link = (data>>6) & 0x1;
		pkts[i].gap_cnt = data & 0x3F;

		data = llc_read(0x62);
		pkts[i].sp = data>>14;
		pkts[i].contender = (data>>11) & 1;
		pkts[i].pwr = (data>>8) & 0x7;
		pkts[i].p0 = (data>>6) & 0x3;
		pkts[i].p1 = (data>>4) & 0x3;
		pkts[i].p2 = (data>>2) & 0x3;
		pkts[i].init = (data>>1) & 0x1;
		pkts[i].more = data & 1;		
	}

	return pkts;
}

// Retrieve this firewire node's 16-bit bus-and-node ID.
int llc_get_node_id()
{
	return llc_read(0x36);
}
	
// Returns 1 if GRF is not empty, 0 if it is
int llc_packet_available()
{
	int dummy;
	int val = llc_read(0x30);
	dummy = llc_read(0x32);
	
	if (val < 0)
		return -1;
	else if (val & 0x0001)
		return 0;
	else
		return 1;
}

// Returns the ack code received for the most
// recent packet sent
int llc_get_last_ack()
{
	int val;
	
	if (llc_read32(0x30, &val) < 0)
		return -1;

	return (val >> 10) & 0xF;
}
